@@ -1,20 +1,21 @@
-/*  Tarlz - Archiver with multimember lzip compression
-    Copyright (C) 2013-2019 Antonio Diaz Diaz.
+/* Tarlz - Archiver with multimember lzip compression
+   Copyright (C) 2013-2020 Antonio Diaz Diaz.
 
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 2 of the License, or
-    (at your option) any later version.
+   This program is free software: you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation, either version 2 of the License, or
+   (at your option) any later version.
 
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
 
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+   You should have received a copy of the GNU General Public License
+   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
+#define max_file_size ( LLONG_MAX - header_size )
 enum { header_size = 512 };
 typedef uint8_t Tar_header[header_size];
 
@@ -70,6 +71,14 @@ inline bool dotdot_at_i( const char * const filename, const int i )
   }
 
 
+inline bool contains_dotdot( const char * const filename )
+  {
+  for( int i = 0; filename[i]; ++i )
+    if( dotdot_at_i( filename, i ) ) return true;
+  return false;
+  }
+
+
 class Resizable_buffer
   {
   char * p;
@@ -104,7 +113,7 @@ class Extended		// stores metadata from/for extended records
   {
   std::string linkpath_;		// these are the real metadata
   std::string path_;
-  unsigned long long file_size_;
+  long long file_size_;			// >= 0 && <= max_file_size
 
   // cached sizes; if full_size_ < 0 they must be recalculated
   mutable long long edsize_;		// extended data size
@@ -137,15 +146,15 @@ public:
 
   const std::string & linkpath() const { return linkpath_; }
   const std::string & path() const { return path_; }
-  unsigned long long file_size() const { return file_size_; }
-  unsigned long long get_file_size_and_reset( const Tar_header header );
+  long long file_size() const { return file_size_; }
+  long long get_file_size_and_reset( const Tar_header header );
 
   void linkpath( const char * const lp ) { linkpath_ = lp; full_size_ = -1; }
   void path( const char * const p ) { path_ = p; full_size_ = -1; }
-  void file_size( const unsigned long long fs )
-    { file_size_ = fs; full_size_ = -1; }
+  void file_size( const long long fs ) { full_size_ = -1;
+    file_size_ = ( fs >= 0 && fs <= max_file_size ) ? fs : 0; }
 
-  unsigned long long full_size() const
+  long long full_size() const
     { if( full_size_ < 0 ) calculate_sizes(); return full_size_; }
 
   bool crc_present() const { return crc_present_; }
@@ -156,6 +165,58 @@ public:
   };
 
 
+class CRC32
+  {
+  uint32_t data[256];		// Table of CRCs of all 8-bit messages.
+
+public:
+  CRC32( const bool castagnoli = false )
+    {
+    const unsigned cpol = 0x82F63B78U;	// CRC32-C  Castagnoli polynomial.
+    const unsigned ipol = 0xEDB88320U;	// IEEE 802.3 Ethernet polynomial.
+    const unsigned poly = castagnoli ? cpol : ipol;
+
+    for( unsigned n = 0; n < 256; ++n )
+      {
+      unsigned c = n;
+      for( int k = 0; k < 8; ++k )
+        { if( c & 1 ) c = poly ^ ( c >> 1 ); else c >>= 1; }
+      data[n] = c;
+      }
+    }
+
+  void update_byte( uint32_t & crc, const uint8_t byte ) const
+    { crc = data[(crc^byte)&0xFF] ^ ( crc >> 8 ); }
+
+  void update_buf( uint32_t & crc, const uint8_t * const buffer,
+                   const int size ) const
+    {
+    uint32_t c = crc;
+    for( int i = 0; i < size; ++i )
+      c = data[(c^buffer[i])&0xFF] ^ ( c >> 8 );
+    crc = c;
+    }
+
+  uint32_t compute_crc( const uint8_t * const buffer, const int size ) const
+    {
+    uint32_t crc = 0xFFFFFFFFU;
+    for( int i = 0; i < size; ++i )
+      crc = data[(crc^buffer[i])&0xFF] ^ ( crc >> 8 );
+    return crc ^ 0xFFFFFFFFU;
+    }
+
+  // Calculates the crc of size bytes except a window of 8 bytes at pos
+  uint32_t windowed_crc( const uint8_t * const buffer, const int pos,
+                         const int size ) const
+    {
+    uint32_t crc = 0xFFFFFFFFU;
+    update_buf( crc, buffer, pos );
+    update_buf( crc, buffer + pos + 8, size - pos - 8 );
+    return crc ^ 0xFFFFFFFFU;
+    }
+  };
+
+
 enum {
   min_dictionary_bits = 12,
   min_dictionary_size = 1 << min_dictionary_bits,
@@ -177,7 +238,7 @@ struct Lzip_header
   {
   uint8_t data[6];			// 0-3 magic bytes
 					//   4 version
-					//   5 coded_dict_size
+					//   5 coded dictionary size
   enum { size = 6 };
 
   bool verify_magic() const
@@ -207,6 +268,10 @@ struct Lzip_header
       sz -= ( sz / 16 ) * ( ( data[5] >> 5 ) & 7 );
     return sz;
     }
+
+  bool verify() const
+    { return verify_magic() && verify_version() &&
+             isvalid_ds( dictionary_size() ); }
   };
 
 
@@ -254,118 +319,60 @@ struct Lzip_trailer
   };
 
 
-class CRC32
+enum Program_mode { m_none, m_append, m_concatenate, m_create, m_delete,
+                    m_diff, m_extract, m_list };
+enum Solidity { no_solid, bsolid, dsolid, asolid, solid };
+class Arg_parser;
+struct Cl_options		// command line options
   {
-  uint32_t data[256];		// Table of CRCs of all 8-bit messages.
+  const Arg_parser & parser;
+  std::string archive_name;
+  long long mtime;
+  Program_mode program_mode;
+  Solidity solidity;
+  int data_size;
+  int debug_level;
+  int filenames;
+  int level;			// compression level, < 0 means uncompressed
+  int num_workers;		// start this many worker threads
+  int out_slots;
+  int owner;
+  int group;
+  bool dereference;
+  bool ignore_ids;
+  bool keep_damaged;
+  bool missing_crc;
+  bool permissive;
+  bool preserve_permissions;
 
-public:
-  CRC32( const bool castagnoli = false )
-    {
-    const unsigned cpol = 0x82F63B78U;	// CRC32-C  Castagnoli polynomial.
-    const unsigned ipol = 0xEDB88320U;	// IEEE 802.3 Ethernet polynomial.
-    const unsigned poly = castagnoli ? cpol : ipol;
-
-    for( unsigned n = 0; n < 256; ++n )
-      {
-      unsigned c = n;
-      for( int k = 0; k < 8; ++k )
-        { if( c & 1 ) c = poly ^ ( c >> 1 ); else c >>= 1; }
-      data[n] = c;
-      }
-    }
-
-  void update_byte( uint32_t & crc, const uint8_t byte ) const
-    { crc = data[(crc^byte)&0xFF] ^ ( crc >> 8 ); }
-
-  void update_buf( uint32_t & crc, const uint8_t * const buffer,
-                   const int size ) const
-    {
-    uint32_t c = crc;
-    for( int i = 0; i < size; ++i )
-      c = data[(c^buffer[i])&0xFF] ^ ( c >> 8 );
-    crc = c;
-    }
-
-  // Calculates the crc of size bytes except a window of 8 bytes at pos
-  uint32_t windowed_crc( const uint8_t * const buffer, const int pos,
-                         const int size ) const
-    {
-    uint32_t crc = 0xFFFFFFFFU;
-    update_buf( crc, buffer, pos );
-    update_buf( crc, buffer + pos + 8, size - pos - 8 );
-    return crc ^ 0xFFFFFFFFU;
-    }
+  Cl_options( const Arg_parser & ap )
+    : parser( ap ), mtime( -1 ), program_mode( m_none ), solidity( bsolid ),
+      data_size( 0 ), debug_level( 0 ), filenames( 0 ), level( 6 ),
+      num_workers( -1 ), out_slots( 64 ), owner( -1 ), group( -1 ),
+      dereference( false ), ignore_ids( false ), keep_damaged( false ),
+      missing_crc( false ), permissive( false ), preserve_permissions( false )
+      {}
   };
 
-extern const CRC32 crc32c;
 
 const char * const bad_magic_msg = "Bad magic number (file not in lzip format).";
 const char * const bad_dict_msg = "Invalid dictionary size in member header.";
 const char * const corrupt_mm_msg = "Corrupt header in multimember file.";
 const char * const trailing_msg = "Trailing data not allowed.";
+const char * const bad_hdr_msg = "Corrupt or invalid tar header.";
+const char * const gblrec_msg = "Error in global extended records.";
+const char * const extrec_msg = "Error in extended records.";
+const char * const mcrc_msg = "Missing CRC in extended records.";
+const char * const end_msg = "Archive ends unexpectedly.";
 const char * const mem_msg = "Not enough memory.";
 const char * const mem_msg2 = "Not enough memory. Try a lower compression level.";
+const char * const fv_msg1 = "Format violation: extended header followed by EOF blocks.";
+const char * const fv_msg2 = "Format violation: extended header followed by global header.";
+const char * const fv_msg3 = "Format violation: consecutive extended headers found.";
+const char * const posix_msg = "This does not look like a POSIX tar archive.";
+const char * const posix_lz_msg = "This does not look like a POSIX tar.lz archive.";
 
-// defined in create.cc
-enum Solidity { no_solid, bsolid, dsolid, asolid, solid };
-extern int cl_owner;
-extern int cl_group;
-extern int cl_data_size;
-extern Solidity solidity;
-bool writeblock_wrapper( const int outfd, const uint8_t * const buffer,
-                         const int size );
-bool write_eof_records( const int outfd, const bool compressed );
-const char * remove_leading_dotslash( const char * const filename,
-                                      const bool dotdot = false );
-bool fill_headers( const char * const filename, Extended & extended,
-                   Tar_header header, unsigned long long & file_size,
-                   const int flag );
-bool block_is_full( const Extended & extended,
-                    const unsigned long long file_size,
-                    unsigned long long & partial_data_size );
-void set_error_status( const int retval );
-int final_exit_status( int retval, const bool show_msg = true );
-unsigned ustar_chksum( const uint8_t * const header );
-bool verify_ustar_chksum( const uint8_t * const header );
-bool has_lz_ext( const std::string & name );
-class Arg_parser;
-int concatenate( std::string archive_name, const Arg_parser & parser,
-                 const int filenames );
-int encode( const std::string & archive_name, const Arg_parser & parser,
-            const int filenames, const int level, const int num_workers,
-            const int out_slots, const int debug_level, const bool append,
-            const bool dereference );
-
-// defined in create_lz.cc
-int encode_lz( const Arg_parser & parser, const int dictionary_size,
-               const int match_len_limit, const int num_workers,
-               const int outfd, const int out_slots, const int debug_level,
-               const bool dereference );
-
-// defined in exclude.cc
-namespace Exclude {
-void add_pattern( const std::string & arg );
-bool excluded( const char * const filename );
-} // end namespace Exclude
-
-// defined in extract.cc
-enum Program_mode { m_none, m_append, m_concatenate, m_create, m_diff,
-                    m_extract, m_list };
-bool block_is_zero( const uint8_t * const buf, const int size );
-bool format_member_name( const Extended & extended, const Tar_header header,
-                         Resizable_buffer & rbuf, const bool long_format );
-bool compare_prefix_dir( const char * const dir, const char * const name );
-bool compare_tslash( const char * const name1, const char * const name2 );
-int readblock( const int fd, uint8_t * const buf, const int size );
-int writeblock( const int fd, const uint8_t * const buf, const int size );
-unsigned long long parse_octal( const uint8_t * const ptr, const int size );
-int decode( const std::string & archive_name, const Arg_parser & parser,
-            const int filenames, const int num_workers, const int debug_level,
-            const Program_mode program_mode, const bool ignore_ids,
-            const bool keep_damaged, const bool missing_crc,
-            const bool permissive );
-
-// defined in list_lz.cc
+// defined in common.cc
 void xinit_mutex( pthread_mutex_t * const mutex );
 void xinit_cond( pthread_cond_t * const cond );
 void xdestroy_mutex( pthread_mutex_t * const mutex );
@@ -375,14 +382,84 @@ void xunlock( pthread_mutex_t * const mutex );
 void xwait( pthread_cond_t * const cond, pthread_mutex_t * const mutex );
 void xsignal( pthread_cond_t * const cond );
 void xbroadcast( pthread_cond_t * const cond );
-bool check_skip_filename( const Arg_parser & parser,
+unsigned long long parse_octal( const uint8_t * const ptr, const int size );
+int readblock( const int fd, uint8_t * const buf, const int size );
+int writeblock( const int fd, const uint8_t * const buf, const int size );
+bool nonempty_arg( const Arg_parser & parser, const int i );
+
+// defined in common_decode.cc
+bool block_is_zero( const uint8_t * const buf, const int size );
+bool format_member_name( const Extended & extended, const Tar_header header,
+                         Resizable_buffer & rbuf, const bool long_format );
+bool show_member_name( const Extended & extended, const Tar_header header,
+                       const int vlevel, Resizable_buffer & rbuf );
+bool check_skip_filename( const Cl_options & cl_opts,
                           std::vector< char > & name_pending,
-                          const char * const filename, const int filenames );
-class Lzip_index;
-int list_lz( const Arg_parser & parser, std::vector< char > & name_pending,
-             const Lzip_index & lzip_index, const int filenames,
-             const int debug_level, const int infd, const int num_workers,
-             const bool missing_crc, const bool permissive );
+                          const char * const filename );
+mode_t get_umask();
+bool make_path( const std::string & name );
+
+// defined in create.cc
+bool copy_file( const int infd, const int outfd, const long long max_size = -1 );
+bool writeblock_wrapper( const int outfd, const uint8_t * const buffer,
+                         const int size );
+bool write_eof_records( const int outfd, const bool compressed );
+const char * remove_leading_dotslash( const char * const filename,
+                                      const bool dotdot = false );
+bool fill_headers( const char * const filename, Extended & extended,
+                   Tar_header header, long long & file_size, const int flag );
+bool block_is_full( const Extended & extended,
+                    const unsigned long long file_size,
+                    unsigned long long & partial_data_size );
+void set_error_status( const int retval );
+int final_exit_status( int retval, const bool show_msg = true );
+unsigned ustar_chksum( const uint8_t * const header );
+bool verify_ustar_chksum( const uint8_t * const header );
+bool has_lz_ext( const std::string & name );
+int concatenate( const Cl_options & cl_opts );
+int encode( Cl_options & cl_opts );
+
+// defined in create_lz.cc
+int encode_lz( const Cl_options & cl_opts, const char * const archive_namep,
+               const int dictionary_size, const int match_len_limit,
+               const int outfd );
+
+// defined in decode.cc
+bool compare_file_type( std::string & estr, std::string & ostr,
+                        const Cl_options & cl_opts,
+                        const Extended & extended, const Tar_header header );
+class Archive_reader_base;
+bool compare_file_contents( std::string & estr, std::string & ostr,
+                            Archive_reader_base & ar, const long long file_size,
+                            const char * const filename, const int infd2 );
+int decode( const Cl_options & cl_opts );
+
+// defined in decode_lz.cc
+struct Archive_descriptor;
+int decode_lz( const Cl_options & cl_opts, const Archive_descriptor & ad,
+               std::vector< char > & name_pending );
+
+// defined in delete.cc
+bool safe_seek( const int fd, const long long pos );
+int tail_copy( const Arg_parser & parser, const Archive_descriptor & ad,
+               std::vector< char > & name_pending, const long long istream_pos,
+               const int outfd, int retval );
+int delete_members( const Cl_options & cl_opts );
+
+// defined in delete_lz.cc
+int delete_members_lz( const Cl_options & cl_opts,
+                       const Archive_descriptor & ad,
+                       std::vector< char > & name_pending, const int outfd );
+
+// defined in exclude.cc
+namespace Exclude {
+void add_pattern( const std::string & arg );
+void clear();
+bool excluded( const char * const filename );
+} // end namespace Exclude
+
+// defined in extended.cc
+extern const CRC32 crc32c;
 
 // defined in lzip_index.cc
 int seek_read( const int fd, uint8_t * const buf, const int size,
@@ -391,12 +468,16 @@ int seek_read( const int fd, uint8_t * const buf, const int size,
 // defined in main.cc
 extern int verbosity;
 struct stat;
-int hstat( const char * const filename, struct stat * const st );
+int hstat( const char * const filename, struct stat * const st,
+           const bool dereference );
 int open_instream( const std::string & name );
-int open_outstream( const std::string & name, const bool create = true );
+int open_outstream( const std::string & name, const bool create = true,
+                    Resizable_buffer * const rbufp = 0 );
 void cleanup_and_fail( const int retval = 1 );	// terminate the program
 void show_error( const char * const msg, const int errcode = 0,
                  const bool help = false );
+void format_file_error( std::string & estr, const char * const filename,
+                        const char * const msg, const int errcode = 0 );
 void show_file_error( const char * const filename, const char * const msg,
                       const int errcode = 0 );
 void internal_error( const char * const msg );
